from os import listdir
from os.path import basename, join, isdir
from typing import List, Literal, Optional, Union

import supervisely.io.env as env
from supervisely._utils import is_development, is_production
from supervisely.api.api import Api
from supervisely.app import DialogWindowError, get_data_dir
from supervisely.app.fastapi.subapp import Application
from supervisely.app.widgets import (
    Button,
    Card,
    Checkbox,
    Container,
    DatasetThumbnail,
    Field,
    FileStorageUpload,
    FileThumbnail,
    FolderThumbnail,
    Input,
    ProjectThumbnail,
    RadioTabs,
    SelectDataset,
    SelectProject,
    SelectWorkspace,
    SlyTqdm,
    Stepper,
    TeamFilesSelector,
    Text,
    Widget,
)
from supervisely.project.project_type import ProjectType
from supervisely.sly_logger import logger
from tqdm import tqdm

PROJECT_TYPES: List[ProjectType] = [
    ProjectType.IMAGES,
    ProjectType.VIDEOS,
    ProjectType.VOLUMES,
    ProjectType.POINT_CLOUDS,
    ProjectType.POINT_CLOUD_EPISODES,
]
DESTINATION_OPTIONS = ["New Project", "Existing Project", "Existing Dataset"]
DATAS_ELECTION_OPTIONS = ["Drag & Drop", "TeamFiles Selector"]
DATA_TYPES = ["folder", "file"]


class Import(Application):
    class Context:
        def __init__(
            self,
            team_id: int,
            workspace_id: int,
            project_id: int,
            dataset_id: int,
            path: str,
            project_name: str,
            progress: SlyTqdm,
            is_on_agent: bool = False,
        ):
            self._team_id = team_id
            if self._team_id is None:
                raise ValueError(f"Team ID is not specified: {self._team_id}")
            if type(self._team_id) is not int:
                raise ValueError(f"Team ID must be 'int': {self._team_id}")

            self._workspace_id = workspace_id
            if self._workspace_id is None:
                raise ValueError(f"Workspace ID is not specified: {self._workspace_id}")
            if type(self._workspace_id) is not int:
                raise ValueError(f"Workspace ID must be 'int': {self._workspace_id}")

            self._project_id = project_id
            if self._project_id is not None and type(self._project_id) is not int:
                raise ValueError(f"Project ID must be 'int': {self._project_id}")

            self._dataset_id = dataset_id
            if self._dataset_id is not None and type(self._dataset_id) is not int:
                raise ValueError(f"Dataset ID must be 'int': {self._dataset_id}")

            self._path = path
            if self._path is None:
                raise ValueError(f"Remote path is not specified: {self._path}")
            if type(self._path) is not str:
                raise ValueError(f"Remote path must be 'str': {self._path}")

            self._project_name = project_name
            if type(self._project_name) is not str:
                raise ValueError(f"Project name must be 'str': {self._project_name}")

            self._progress = progress

            self._is_on_agent = is_on_agent
            if type(self._is_on_agent) is not bool:
                raise ValueError(f"Flag 'is_on_agent' must be 'bool': {self._is_on_agent}")

        def __str__(self):
            return (
                f"Team ID: {self._team_id}\n"
                f"Workspace ID: {self._workspace_id}\n"
                f"Project ID: {self._project_id}\n"
                f"Dataset ID: {self._dataset_id}\n"
                f"Path: {self._path}\n"
                f"Project name: {self._project_name}\n"
                f"Is on agent: {self._is_on_agent}\n"
            )

        @property
        def team_id(self) -> int:
            return self._team_id

        @property
        def workspace_id(self) -> int:
            return self._workspace_id

        @property
        def project_id(self) -> int:
            return self._project_id

        @property
        def dataset_id(self) -> int:
            return self._dataset_id

        @property
        def path(self) -> str:
            return self._path

        @property
        def project_name(self) -> str:
            return self._project_name

        @property
        def progress(self) -> SlyTqdm or tqdm:
            return self._progress

        @property
        def is_on_agent(self) -> bool:
            return self._is_on_agent

    def __init__(
        self,
        allowed_project_types: List[ProjectType] = None,
        allowed_destination_options: List[
            Literal["New Project", "Existing Project", "Existing Dataset"]
        ] = None,
        allowed_data_type: Literal["folder", "file"] = None,
    ):
        if allowed_project_types is None:
            self._allowed_project_types = PROJECT_TYPES
        else:
            for option in allowed_project_types:
                if option not in PROJECT_TYPES:
                    raise ValueError(f"Project type must be one of {PROJECT_TYPES}: {option}")
            if len(allowed_project_types) > 5:
                raise ValueError(
                    f"Maximum 5 project types can be specified: {allowed_project_types}"
                )
            self._allowed_project_types: List[ProjectType] = allowed_project_types

        if allowed_destination_options is None:
            self._allowed_destination_options = DESTINATION_OPTIONS
        else:
            for option in allowed_destination_options:
                if option not in DESTINATION_OPTIONS:
                    raise ValueError(f"Destination must be one of {DESTINATION_OPTIONS}: {option}")
            if len(allowed_destination_options) > 3:
                raise ValueError(
                    f"Maximum 3 destinations can be specified: {allowed_destination_options}"
                )
            self._allowed_destination_options: Literal[
                "New Project", "Existing Project", "Existing Dataset"
            ] = allowed_destination_options

        if allowed_data_type is None:
            self._allowed_data_type = None
        else:
            if allowed_data_type not in DATA_TYPES:
                raise ValueError(
                    f"Data type must be one of {DATA_TYPES} or 'None' to use both: {allowed_data_type}"
                )

            self._allowed_data_type = allowed_data_type

        self.__layout = None
        api = Api.from_env()

        self.__step_one_btn = Button("Next", icon="zmdi zmdi-check")

        self.__step_one_text = Text()
        self.__step_one_text.hide()

        self.__step_two_btn = Button("Next", icon="zmdi zmdi-check")
        self.__step_two_btn.disable()

        self.__step_three_btn = Button("Next", icon="zmdi zmdi-check")
        self.__step_three_btn.disable()

        self.__step_three_text = Text()
        self.__step_three_text.hide()

        self.__start_button = Button("Start Import")
        self.__start_button.disable()

        self.__output_text = Text()
        self.__output_text.hide()

        self.__output_project_tabs = None
        self.__input_data_tabs = None
        # self.__step_four_btn = Button("Next")

        ####################
        # PROJECT SELECTOR #
        ####################

        self.__input_mode = None

        # Launch from Dataset
        if (
            env.project_id(raise_not_found=False) is not None
            and env.dataset_id(raise_not_found=False) is not None
        ):
            project_info = api.project.get_info_by_id(env.project_id())
            dataset_info = api.dataset.get_info_by_id(env.dataset_id())
            self.__input_thumbnail = DatasetThumbnail(
                project_info=project_info, dataset_info=dataset_info
            )
            self.__project_selector_card_widgets = Container(widgets=[self.__input_thumbnail])
            self.__input_mode = "dataset"

        # Launch from Project
        elif env.project_id(raise_not_found=False) is not None:
            project_info = api.project.get_info_by_id(env.project_id())
            self.__input_thumbnail = ProjectThumbnail(info=project_info)
            self.__project_selector_card_widgets = Container(widgets=[self.__input_thumbnail])
            self.__input_mode = "project"
        # Launch from Ecosystem
        else:
            self.__output_workspace_selector = SelectWorkspace(
                default_id=env.workspace_id(), team_id=env.team_id()
            )
            self.__output_new_project_name = Input(value="My project")
            self.__output_new_project = Container(
                widgets=[
                    self.__output_workspace_selector,
                    Text("Project name"),
                    self.__output_new_project_name,
                ]
            )
            self.__output_project_selector = SelectProject(
                allowed_types=self._allowed_project_types
            )
            self.__output_dataset_selector = SelectDataset(
                allowed_project_types=self._allowed_project_types
            )
            destination_widgets_map = {
                "New Project": self.__output_new_project,
                "Existing Project": self.__output_project_selector,
                "Existing Dataset": self.__output_dataset_selector,
            }

            if len(self._allowed_destination_options) > 1:
                self.__output_project_tabs = RadioTabs(
                    titles=self._allowed_destination_options,
                    contents=[
                        destination_widgets_map[option]
                        for option in self._allowed_destination_options
                    ],
                )
                self.__project_selector_card_widgets = Container(
                    widgets=[self.__output_project_tabs, self.__step_three_text]
                )
            else:
                self.__project_selector_card_widgets = Container(
                    widgets=[
                        destination_widgets_map[self._allowed_destination_options[0]],
                        self.__step_three_text,
                    ]
                )
            self.__input_mode = "ecosystem"

        ####################
        #  FILES SELECTOR  #
        ####################

        # Launch from File
        if is_production():
            if env.file(raise_not_found=False) is not None:
                file_info = api.file.get_info_by_path(team_id=env.team_id(), remote_path=env.file())
                self.__input_thumbnail = FileThumbnail(info=file_info)
                self.__files_selector_card_widgets = Container(widgets=[self.__input_thumbnail])
                if self.__input_mode is None or self.__input_mode == "ecosystem":
                    self.__input_mode = "file"
                self.__step_two_btn.enable()

            # Launch from Folder
            elif env.folder(raise_not_found=False) is not None:
                import tempfile

                temp_file = tempfile.NamedTemporaryFile(suffix=".tmp", delete=False)
                temp_file.close()
                file_path = temp_file.name
                remote_path = join(env.folder(), basename(file_path))
                file_info = api.file.upload(team_id=env.team_id(), src=file_path, dst=remote_path)
                self.__input_thumbnail = FolderThumbnail(info=file_info)
                api.file.remove(team_id=env.team_id(), path=remote_path)
                self.__files_selector_card_widgets = Container(widgets=[self.__input_thumbnail])
                if self.__input_mode is None or self.__input_mode == "ecosystem":
                    self.__input_mode = "folder"
                self.__step_two_btn.enable()

            # Launch from Ecosystem
            else:
                self.__input_file_selector = TeamFilesSelector(
                    team_id=env.team_id(),
                    multiple_selection=False,
                    selection_file_type=self._allowed_data_type,
                    max_height=300,
                )

                app_name = env.app_name(raise_not_found=False)
                if app_name is None:
                    app_name = "import-template"
                else:
                    app_name = app_name.lower().replace(" ", "-")

                if env.task_id(raise_not_found=False) is None:
                    self.__storage_upload_path = f"/import/{app_name}/debug/"
                else:
                    self.__storage_upload_path = f"/import/{app_name}/{env.task_id()}/"

                self.__input_drag_n_drop = FileStorageUpload(
                    team_id=env.team_id(), path=self.__storage_upload_path
                )
                self.__input_data_tabs = RadioTabs(
                    titles=["Drag & Drop", "TeamFiles Selector"],
                    contents=[
                        self.__input_drag_n_drop,
                        self.__input_file_selector,
                    ],
                )
                self.__files_selector_card_widgets = Container(
                    widgets=[self.__input_data_tabs, self.__step_one_text]
                )
                if self.__input_mode is None:
                    self.__input_mode = "ecosystem"

        # if is_development():
        else:
            data_dir = get_data_dir()
            self.__input_file_text = Text()
            if not isdir(data_dir):
                self.__input_file_text.set(
                    f"Please create data folder specified in 'local.env' and put data for import there",
                    "warning",
                )
                self.__step_one_btn.disable()

            if len(listdir(data_dir)) == 0:
                self.__input_file_text.set(
                    f"Please put the data in the data folder: {data_dir}", "warning"
                )
                self.__step_one_btn.disable()

            else:
                self.__input_file_text.set(f"Data folder: {data_dir}", "success")

            self.__files_selector_card_widgets = Container(widgets=[self.__input_file_text])

        ############
        # SETTINGS #
        #############

        self.__remove_source_files_checkbox = Checkbox(
            "Remove source files after import", checked=True
        )

        if is_development():
            self.__remove_source_files_checkbox.uncheck()
            self.__remove_source_files_checkbox.disable()

        self.__settings_card_widgets = Container(widgets=[self.__remove_source_files_checkbox])
        custom_settings_container = self.add_custom_settings()
        if custom_settings_container is not None:
            if not isinstance(custom_settings_container, Widget):
                raise ValueError("Custom settings must return a Widget")
            if type(custom_settings_container) is not Container:
                custom_settings_container = Container(widgets=[custom_settings_container])
            if isinstance(custom_settings_container, Widget):
                custom_field = Field(content=custom_settings_container, title="Custom Settings")
                self.__settings_card_widgets._widgets.append(custom_field)

        self.__settings_card = Card(
            title="Import Settings",
            description="Configure import settings",
            content=self.__settings_card_widgets,
        )

        ##########
        # OUTPUT #
        ##########

        self.__output_project_thumbnail = ProjectThumbnail()
        self.__output_project_thumbnail.hide()
        self.__import_progress = SlyTqdm()
        self.__output_card_container = Container(
            widgets=[
                self.__output_project_thumbnail,
                self.__output_text,
                self.__start_button,
                self.__import_progress,
            ]
        )
        self.__output_card = Card(
            title="Output",
            description="Link to access result project will appear below after successful import",
            content=self.__output_card_container,
        )

        ##########
        # LAYOUT #
        ##########
        if is_production():
            input_files_card_desc = (
                "Select data to import. You can select data from Team Files or use Drag & Drop"
            )
            local_debug_msg = ""
        else:
            input_files_card_desc = "Application was launched in local debug mode. Data folder is specified in local.env file"
            local_debug_msg = " in local debug mode"

        if self.__input_mode == "project":
            self.__input_project_card = Card(
                title="Destination Project",
                description=f"Application was launched from the context menu of the project{local_debug_msg}",
                content=self.__project_selector_card_widgets,
            )
            self.__input_files_card = Card(
                title="Select Data",
                description=f"{input_files_card_desc}",
                content=self.__files_selector_card_widgets,
            )

            self.__input_project_container = Container(widgets=[self.__input_project_card])
            self.__files_selector_container = Container(
                widgets=[self.__input_files_card, self.__step_one_btn]
            )
            self.__settings_container = Container(
                widgets=[self.__settings_card, self.__step_two_btn]
            )

            self.__layout = Stepper(
                widgets=[
                    self.__input_project_container,
                    self.__files_selector_container,
                    self.__settings_container,
                    self.__output_card,
                ],
                active_step=2,
            )

            self.__settings_card.lock(message="Select data to import to unlock this card")
            self.__output_card.lock(message="Configure import settings to unlock this card")

        elif self.__input_mode == "dataset":
            self.__input_project_card = Card(
                title="Destination Dataset",
                description=f"Application was launched from the context menu of the dataset{local_debug_msg}",
                content=self.__project_selector_card_widgets,
            )
            self.__input_files_card = Card(
                title="Select Data",
                description=f"{input_files_card_desc}",
                content=self.__files_selector_card_widgets,
            )

            self.__input_project_container = Container(widgets=[self.__input_project_card])
            self.__files_selector_container = Container(
                widgets=[self.__input_files_card, self.__step_one_btn]
            )
            self.__settings_container = Container(
                widgets=[self.__settings_card, self.__step_two_btn]
            )

            self.__layout = Stepper(
                widgets=[
                    self.__input_project_container,
                    self.__files_selector_container,
                    self.__settings_container,
                    self.__output_card,
                ],
                active_step=2,
            )
            self.__settings_card.lock(message="Select data to import to unlock this card")
            self.__output_card.lock(message="Configure import settings to unlock this card")

        elif self.__input_mode == "file":
            self.__input_project_card = Card(
                title="Destination Project",
                description="Select where do you want to import your data",
                content=self.__project_selector_card_widgets,
            )
            self.__input_files_card = Card(
                title="Input File",
                description=f"Application was launched from the context menu of the file{local_debug_msg}",
                content=self.__files_selector_card_widgets,
            )

            self.__files_selector_container = Container(widgets=[self.__input_files_card])

            self.__settings_container = Container(
                widgets=[self.__settings_card, self.__step_two_btn]
            )
            self.__input_project_container = Container(
                widgets=[self.__input_project_card, self.__step_three_btn]
            )

            self.__layout = Stepper(
                widgets=[
                    self.__files_selector_container,
                    self.__settings_container,
                    self.__input_project_container,
                    self.__output_card,
                ],
                active_step=2,
            )
            self.__input_project_card.lock(message="Configure import settings to unlock this card")
            self.__output_card.lock(message="Select destination project to unlock this card")

        elif self.__input_mode == "folder":
            self.__input_project_card = Card(
                title="Destination Project",
                description="Select where do you want to import your data",
                content=self.__project_selector_card_widgets,
            )
            self.__input_files_card = Card(
                title="Input Folder",
                description=f"Application was launched from the context menu of the folder{local_debug_msg}",
                content=self.__files_selector_card_widgets,
            )

            self.__files_selector_container = Container(widgets=[self.__input_files_card])

            self.__settings_container = Container(
                widgets=[self.__settings_card, self.__step_two_btn]
            )
            self.__input_project_container = Container(
                widgets=[self.__input_project_card, self.__step_three_btn]
            )

            self.__layout = Stepper(
                widgets=[
                    self.__files_selector_container,
                    self.__settings_container,
                    self.__input_project_container,
                    self.__output_card,
                ],
                active_step=2,
            )
            self.__input_project_card.lock(message="Configure import settings to unlock this card")
            self.__output_card.lock(message="Select destination project to unlock this card")

        else:
            self.__input_project_card = Card(
                title="Destination Project",
                description="Select where do you want to import your data",
                content=self.__project_selector_card_widgets,
            )
            self.__input_files_card = Card(
                title="Select Data",
                description=f"{input_files_card_desc}",
                content=self.__files_selector_card_widgets,
            )

            self.__files_selector_container = Container(
                widgets=[self.__input_files_card, self.__step_one_btn]
            )
            self.__settings_container = Container(
                widgets=[self.__settings_card, self.__step_two_btn]
            )
            self.__input_project_container = Container(
                widgets=[self.__input_project_card, self.__step_three_btn]
            )

            self.__layout = Stepper(
                widgets=[
                    self.__files_selector_container,
                    self.__settings_container,
                    self.__input_project_container,
                    self.__output_card,
                ],
                active_step=1,
            )
            self.__settings_card.lock("Select data to import to unlock this card")
            self.__input_project_card.lock(message="Configure import settings to unlock this card")
            self.__output_card.lock(message="Select destination project to unlock this card")
        super().__init__(layout=self.__layout, hot_reload=False)

    def __generate_output_text(self, api: Api):
        team_id, workspace_id, project_id, dataset_id, project_name = self.__get_destination(api)
        if is_production():
            path = self.__get_remote_path()
        else:
            path = get_data_dir()
        team_name = api.team.get_info_by_id(team_id).name
        workspace_name = api.workspace.get_info_by_id(workspace_id).name

        dst_status = "new"
        if project_id is not None:
            project_name = api.project.get_info_by_id(project_id).name
            dst_status = "existing"

        if self.__remove_source_files_checkbox.is_checked():
            remove_source_message = "Temporary data will be removed after a successful import in order to save disk space."
        else:
            remove_source_message = (
                "Data won't be deleted and will remain accessible after a successful import."
            )

        tf_msg = f"<b>You can access data in Team Files by following path:</b> {path}<br>"
        local_m_msg = f"<b>You can access data on local machine by following path:</b> {path}<br>"
        if project_id is not None and dataset_id is not None:
            dataset_name = api.dataset.get_info_by_id(dataset_id).name
            ds_msg = (
                f"<b>Data will be imported to {dst_status} dataset:</b> {dataset_name} (team: {team_name}, workspace: {workspace_name}, project: {project_name})<br>"
                f"<b>{remove_source_message}</b>"
            )

            if is_production():
                output_text = f"{tf_msg}" f"{ds_msg}"
            else:
                output_text = f"{local_m_msg}" f"{ds_msg}"
        else:
            proj_msg = (
                f"<b> Data will be imported to {dst_status} project:</b> {project_name} (team: {team_name}, workspace: {workspace_name})<br>"
                f"<b>{remove_source_message}</b>"
            )

            if is_production():
                output_text = f"{tf_msg}" f"{proj_msg}"
            else:
                output_text = f"{local_m_msg}" f"{proj_msg}"

        return output_text

    def __get_destination(self, api: Api):
        team_id = env.team_id()
        workspace_id = env.workspace_id()
        project_id = None
        dataset_id = None
        project_name = None
        if is_production():
            if self.__input_mode == "project":
                project_id = env.project_id()
            elif self.__input_mode == "dataset":
                project_id = env.project_id()
                dataset_id = env.dataset_id()
            elif self.__input_mode in ["ecosystem", "file", "folder"]:
                project_mode = self.__output_project_tabs.get_active_tab()
                if project_mode == "New Project":
                    team_id = self.__output_workspace_selector._team_selector.get_selected_id()
                    workspace_id = self.__output_workspace_selector.get_selected_id()
                    project_name = self.__output_new_project_name.get_value()
                elif project_mode == "Existing Project":
                    team_id = (
                        self.__output_project_selector._ws_selector._team_selector.get_selected_id()
                    )
                    workspace_id = self.__output_project_selector._ws_selector.get_selected_id()
                    project_id = self.__output_project_selector.get_selected_id()
                    dataset_id = None
                elif project_mode == "Existing Dataset":
                    team_id = (
                        self.__output_dataset_selector._project_selector._ws_selector._team_selector.get_selected_id()
                    )
                    workspace_id = (
                        self.__output_dataset_selector._project_selector._ws_selector.get_selected_id()
                    )
                    project_id = self.__output_dataset_selector._project_selector.get_selected_id()
                    dataset_id = self.__output_dataset_selector.get_selected_id()

        if project_name is None and project_id is not None:
            project_name = api.project.get_info_by_id(project_id).name

        return team_id, workspace_id, project_id, dataset_id, project_name

    def __get_remote_path(self):
        if self.__input_mode == "file":
            path = env.file()
        elif self.__input_mode == "folder":
            path = env.folder()
        elif self.__input_mode in ["ecosystem", "project", "dataset"]:
            data_mode = self.__input_data_tabs.get_active_tab()
            if data_mode == "TeamFiles Selector":
                path = self.__input_file_selector.get_selected_paths()[0]
            elif data_mode == "Drag & Drop":
                path = self.__storage_upload_path

        return path

    def __get_context(self, api: Api) -> Context:
        team_id = env.team_id()
        workspace_id = env.workspace_id()
        project_id = None
        dataset_id = None
        path = None
        project_name = ""
        progress = self.__import_progress
        is_on_agent = False

        if is_production():
            team_id, workspace_id, project_id, dataset_id, project_name = self.__get_destination(
                api
            )

            data_dir = get_data_dir()
            if self.__input_mode == "file":
                path = env.file()
                local_save_path = join(data_dir, basename(path.rstrip("/")))
                api.file.download(team_id=team_id, remote_path=path, local_save_path=path)
                path = local_save_path
            elif self.__input_mode == "folder":
                path = env.folder()
                local_save_path = join(data_dir, basename(path.rstrip("/")))
                api.file.download_directory(
                    team_id=team_id, remote_path=path, local_save_path=local_save_path
                )
                path = local_save_path
            elif self.__input_mode in ["ecosystem", "project", "dataset"]:
                data_mode = self.__input_data_tabs.get_active_tab()
                if data_mode == "TeamFiles Selector":
                    paths = self.__input_file_selector.get_selected_paths()

                    data_path = self.__input_file_selector.get_selected_paths()[0]
                    data_type = self.__input_file_selector.get_selected_items()[0]["type"]

                    path = join(data_dir, basename(data_path.rstrip("/")))
                    if data_type == "file":
                        api.file.download(
                            team_id=team_id, remote_path=data_path, local_save_path=path
                        )
                    else:
                        api.file.download_directory(
                            team_id=team_id, remote_path=data_path, local_save_path=path
                        )
                elif data_mode == "Drag & Drop":
                    paths = self.__input_drag_n_drop.get_uploaded_paths()
                    for path in paths:
                        local_save_path = join(
                            data_dir, path.replace(self.__storage_upload_path, "")
                        )
                        api.file.download(
                            team_id=team_id, remote_path=path, local_save_path=local_save_path
                        )
                    path = data_dir

            return Import.Context(
                team_id=team_id,
                workspace_id=workspace_id,
                project_id=project_id,
                dataset_id=dataset_id,
                path=get_data_dir(),
                project_name=project_name,
                progress=progress,
                is_on_agent=is_on_agent,
            )

        # if is_development():
        else:
            data_dir = get_data_dir()

            if not isdir(data_dir):
                raise ValueError(
                    "Please create data folder specified in 'local.env' and put data for import there"
                )

            if len(listdir(data_dir)) == 0:
                raise ValueError(f"Please put the data in the data folder: {data_dir}")

            # file = env.file(raise_not_found=False)
            # folder = env.folder(raise_not_found=False)

            # if file is not None and folder is not None:
            #     raise KeyError(
            #         "Both FILE and FOLDER envs are defined, but only one is allowed for the import"
            #     )

            # if file is None and folder is None:
            #     raise KeyError(
            #         "One of the environment variables has to be defined for the import app: FILE or FOLDER"
            #     )

            # path = folder
            # if file is not None:
            #     path = file

            is_on_agent = False
            # if path is not None:
            #     is_on_agent = api.file.is_on_agent(path)

            project_id = env.project_id(raise_not_found=False)
            dataset_id = env.dataset_id(raise_not_found=False)
            if project_id is not None:
                project = api.project.get_info_by_id(id=project_id)
                if project is None:
                    raise ValueError(
                        f"Project with ID: '{project_id}' is not found or either archived"
                    )
                project_id = project.id
                logger.info(f"Importing to existing Project: id={project.id}, name={project.name}")

            if dataset_id is not None:
                dataset = api.dataset.get_info_by_id(id=dataset_id)
                if dataset is None:
                    raise ValueError(
                        f"Dataset with ID: '{dataset_id}' is not found or either archived"
                    )
                if project is not None and dataset not in api.dataset.get_list(
                    project_id=project.id
                ):
                    raise ValueError(
                        f"Dataset {dataset.name}(ID {dataset.id}) "
                        f"does not belong to project {project.name} (ID {project.id})."
                    )
                dataset_id = dataset.id
                logger.info(f"Importing to existing Dataset: id={dataset.id}, name={dataset.name}")

            return Import.Context(
                team_id=team_id,
                workspace_id=workspace_id,
                project_id=project_id,
                dataset_id=dataset_id,
                path=get_data_dir(),
                project_name=project_name,
                progress=progress,
                is_on_agent=is_on_agent,
            )

    def add_custom_settings(self) -> List[Widget]:
        # implement your own method for import settings
        return None

    def process(self, context: Context) -> Optional[Union[int, None]]:
        # implement your own method for import
        raise NotImplementedError()

    def run(self):
        api = Api.from_env()

        # if is_development():
        #     context = self.__get_context(api)
        #     project_id = self.process(context)
        #     info = api.project.get_info_by_id(project_id)
        #     logger.info(f"Result project: id={info.id}, name={info.name}")

        # if is_production() is True:
        if self.__input_data_tabs is not None:

            @self.__input_data_tabs.value_changed
            def hide_message_files(value):
                self.__step_one_text.hide()

        if self.__output_project_tabs is not None:

            @self.__output_project_tabs.value_changed
            def hide_message_project(value):
                self.__step_three_text.hide()

        @self.__step_one_btn.click
        def finish_step_one():
            if self.__input_data_tabs is not None:
                if self.__input_data_tabs.get_active_tab() == "TeamFiles Selector":
                    if len(self.__input_file_selector.get_selected_items()) == 0:
                        self.__step_one_text.show()
                        self.__step_one_text.set(text="Select a data to import", status="error")
                        return
                elif self.__input_data_tabs.get_active_tab() == "Drag & Drop":
                    if len(self.__input_drag_n_drop.get_uploaded_paths()) == 0:
                        self.__step_one_text.show()
                        self.__step_one_text.set(text="Drag & Drop data to import", status="error")
                        return

            if self.__step_one_btn.text == "Next":
                self.__step_one_text.hide()

                if self.__input_mode in ["dataset", "project"]:
                    self.__input_project_card.unlock()
                    self.__input_files_card.lock(
                        message="Press the button to restart from this step"
                    )
                    self.__layout.set_active_step(3)

                elif self.__input_mode in ["file", "folder"]:
                    self.__input_project_card.unlock()
                    self.__layout.set_active_step(3)

                elif self.__input_mode == "ecosystem":
                    self.__input_files_card.lock(
                        message="Press the button to restart from this step"
                    )
                    self.__input_project_card.lock()
                    self.__layout.set_active_step(2)

                self.__step_one_btn.text = "Restart"
                self.__step_one_btn.icon = "zmdi zmdi-rotate-left"

                self.__settings_card.unlock()
                self.__step_two_btn.enable()

            elif self.__step_one_btn.text == "Restart":
                self.__step_one_text.hide()

                if self.__input_mode in ["dataset", "project"]:
                    self.__input_project_card.unlock()
                    self.__input_files_card.unlock()
                    self.__output_card.lock(message="Configure import settings to unlock this card")
                    self.__layout.set_active_step(2)
                elif self.__input_mode in ["file", "folder"]:
                    self.__input_project_card.unlock()
                    self.__layout.set_active_step(2)
                    self.__output_card.lock(
                        message="Select destination project to unlock this card"
                    )
                elif self.__input_mode == "ecosystem":
                    self.__input_files_card.unlock()
                    self.__input_project_card.lock(
                        message="Configure import settings to unlock this card"
                    )
                    self.__layout.set_active_step(1)
                    self.__output_card.lock(
                        message="Select destination project to unlock this card"
                    )

                self.__step_two_btn.text = "Next"
                self.__step_two_btn.icon = "zmdi zmdi-check"
                self.__step_three_btn.text = "Next"
                self.__step_three_btn.icon = "zmdi zmdi-check"

                self.__step_one_btn.text = "Next"
                self.__step_one_btn.icon = "zmdi zmdi-check"
                self.__step_two_btn.disable()
                self.__step_three_btn.disable()
                self.__settings_card.lock(message="Select data to import to unlock this card")

            self.__output_text.hide()
            self.__start_button.disable()

        @self.__step_two_btn.click
        def finish_step_two():
            if self.__step_two_btn.text == "Next":
                if self.__input_mode in ["dataset", "project"]:
                    self.__input_files_card.lock()
                    self.__settings_card.lock(message="Press the button to restart from this step")

                    output_text = self.__generate_output_text(api)
                    self.__output_text.set(
                        text=output_text,
                        status="info",
                    )
                    self.__output_text.show()
                    self.__start_button.enable()
                    self.__output_card.unlock()
                    self.__layout.set_active_step(4)

                elif self.__input_mode in ["file", "folder"]:
                    self.__input_project_card.unlock()
                    self.__settings_card.lock(message="Press the button to restart from this step")
                    self.__layout.set_active_step(3)
                elif self.__input_mode == "ecosystem":
                    self.__input_files_card.lock()
                    self.__input_project_card.unlock()
                    self.__settings_card.lock(message="Press the button to restart from this step")
                    self.__layout.set_active_step(3)

                self.__step_two_btn.text = "Restart"
                self.__step_two_btn.icon = "zmdi zmdi-rotate-left"
                self.__step_three_btn.enable()

            elif self.__step_two_btn.text == "Restart":
                if self.__input_mode in ["dataset", "project"]:
                    self.__input_files_card.lock()
                    self.__settings_card.unlock()
                    self.__layout.set_active_step(3)
                elif self.__input_mode in ["file", "folder"]:
                    self.__settings_card.unlock()
                    self.__input_project_card.lock(
                        message="Configure import settings to unlock this card"
                    )
                    self.__layout.set_active_step(2)
                elif self.__input_mode == "ecosystem":
                    self.__input_files_card.lock()
                    self.__settings_card.unlock()
                    self.__input_project_card.lock(
                        message="Configure import settings to unlock this card"
                    )
                    self.__layout.set_active_step(2)

                self.__step_three_btn.text = "Next"
                self.__step_three_btn.icon = "zmdi zmdi-check"
                self.__step_three_btn.disable()
                self.__output_card.lock()

                self.__output_text.hide()
                self.__start_button.disable()
                self.__step_two_btn.text = "Next"
                self.__step_two_btn.icon = "zmdi zmdi-check"

        @self.__step_three_btn.click
        def finish_step_three():
            if (
                self.__output_project_tabs is not None
                and len(self._allowed_destination_options) > 1
            ):
                if self.__output_project_tabs.get_active_tab() == "New Project":
                    if self.__output_new_project_name.get_value() == "":
                        self.__step_three_text.show()
                        self.__step_three_text.set(text="Enter a project name", status="error")
                        return
                elif self.__output_project_tabs.get_active_tab() == "Existing Project":
                    if self.__output_project_selector.get_selected_id() is None:
                        self.__step_three_text.show()
                        self.__step_three_text.set(
                            text="Select a project to import to", status="error"
                        )
                        return
                elif self.__output_project_tabs.get_active_tab() == "Existing Dataset":
                    if (
                        self.__output_dataset_selector.get_selected_id() is None
                        or type(self.__output_dataset_selector.get_selected_id()) is not int
                    ):
                        self.__step_three_text.show()
                        self.__step_three_text.set(
                            text="Select a dataset to import to", status="error"
                        )
                        return

            else:
                if self._allowed_destination_options == ["New Project"]:
                    if self.__output_new_project_name.get_value() == "":
                        self.__step_three_text.show()
                        self.__step_three_text.set(text="Enter a project name", status="error")
                        return
                elif self._allowed_destination_options == ["Existing Project"]:
                    if self.__output_project_selector.get_selected_id() is None:
                        self.__step_three_text.show()
                        self.__step_three_text.set(
                            text="Select a project to import to", status="error"
                        )
                        return
                elif self._allowed_destination_options == ["Existing Dataset"]:
                    if (
                        self.__output_dataset_selector.get_selected_id() is None
                        or len(self.__output_dataset_selector.get_selected_id()) == 0
                    ):
                        self.__step_three_text.show()
                        self.__step_three_text.set(
                            text="Select a dataset to import to", status="error"
                        )
                        return

            if self.__step_three_btn.text == "Next":
                self.__step_three_text.hide()
                if self.__input_mode in ["dataset", "project"]:
                    self.__input_files_card.unlock()
                elif self.__input_mode in ["file", "folder"]:
                    self.__settings_card.lock()
                    self.__input_project_card.lock(
                        message="Press the button to restart from this step"
                    )
                elif self.__input_mode == "ecosystem":
                    self.__settings_card.lock()
                    self.__input_project_card.lock(
                        message="Press the button to restart from this step"
                    )

                output_text = self.__generate_output_text(api)
                self.__output_text.set(
                    text=output_text,
                    status="info",
                )
                self.__output_text.show()

                self.__start_button.enable()
                self.__layout.set_active_step(4)

                self.__step_three_btn.text = "Restart"
                self.__step_three_btn.icon = "zmdi zmdi-rotate-left"
                self.__output_card.unlock()

            elif self.__step_three_btn.text == "Restart":
                self.__step_three_text.hide()
                if self.__input_mode in ["dataset", "project"]:
                    self.__input_files_card.unlock()
                    self.__output_card.lock(message="Configure import settings to unlock this card")
                elif self.__input_mode in ["file", "folder"]:
                    self.__settings_card.lock()
                    self.__input_project_card.unlock()
                    self.__output_card.lock(
                        message="Select destination project to unlock this card"
                    )
                elif self.__input_mode == "ecosystem":
                    self.__input_project_card.unlock()
                    self.__settings_card.lock()
                    self.__output_card.lock(
                        message="Select destination project to unlock this card"
                    )

                self.__layout.set_active_step(3)

                self.__output_text.hide()
                self.__start_button.disable()
                self.__step_three_btn.text = "Next"
                self.__step_three_btn.icon = "zmdi zmdi-check"

        @self.__start_button.click
        def start_import():
            try:
                self.__step_one_btn.disable()
                self.__step_two_btn.disable()
                self.__step_three_btn.disable()
                self.__output_text.hide()
                context = self.__get_context(api)
                self.__import_progress.show()

                project_id = self.process(context)
                if type(project_id) is int:
                    info = api.project.get_info_by_id(project_id)
                    if is_production() and self.__remove_source_files_checkbox.is_checked():
                        try:
                            dd_paths = self.__input_drag_n_drop.get_uploaded_paths()
                            tf_paths = self.__input_file_selector.get_selected_paths()
                            paths = dd_paths + tf_paths
                            for path in paths:
                                api.file.remove(team_id=env.team_id(), path=path)
                            self.__output_text.set(
                                text="Source files have been successfully removed",
                                status="success",
                            )
                            self.__output_text.show()
                        except Exception as e:
                            self.__output_text.set(
                                text=f"Couldn't remove source files<br>Error: {e}",
                                status="error",
                            )
                            self.__output_text.show()

                    logger.info(f"Result project: id={info.id}, name={info.name}")
                    self.__output_project_thumbnail.set(info)
                    self.__output_text.set("Import completed successfully", status="success")

                    self.__output_project_thumbnail.show()
                    self.__output_text.show()
                else:
                    self.__step_one_btn.enable()
                    self.__step_two_btn.enable()
                    self.__step_three_btn.enable()
                    raise DialogWindowError(
                        title="Import error",
                        description=f"Error: '{project_id}' is not a valid project id",
                    )

                self.__start_button.disable()
            except Exception as e:
                self.__import_progress.hide()
                self.__output_project_thumbnail.hide()
                self.__step_one_btn.enable()
                self.__step_two_btn.enable()
                self.__step_three_btn.enable()
                self.__start_button.enable()
                self.__start_button.loading = False
                raise DialogWindowError(title="Import error", description=f"Error: {str(e)}")



# coding: utf-8

# docs
from __future__ import annotations
import numpy as np
import base64
import gzip
import nrrd
import tempfile
from typing import Optional, Union, List, Tuple, Dict, Literal
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.constants import (
    SPACE_ORIGIN,
    DATA,
    GEOMETRY_SHAPE,
    GEOMETRY_TYPE,
    LABELER_LOGIN,
    UPDATED_AT,
    CREATED_AT,
    ID,
    CLASS_ID,
    MASK_3D,
)
from supervisely._utils import unwrap_if_numpy
from supervisely.io.json import JsonSerializable
from supervisely.io.fs import remove_dir, get_file_name
from supervisely import logger


if not hasattr(np, "bool"):
    np.bool = np.bool_


class PointVolume(JsonSerializable):
    """
    PointVolume (x, y, z) determines position of Mask3D. It locates the first sample.
    :class:`PointVolume<PointVolume>` object is immutable.

    :param x: Position of PointVolume object on X-axis.
    :type x: int or float
    :param y: Position of PointVolume object on Y-axis.
    :type y: int or float
    :param z: Position of PointVolume object on Z-axis.
    :type z: int or float
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        x = 100
        y = 200
        z = 2
        loc = sly.PointVolume(x, y, z)
    """

    def __init__(self, x: Union[int, float], y: Union[int, float], z: Union[int, float]):
        self._x = round(unwrap_if_numpy(x))
        self._y = round(unwrap_if_numpy(y))
        self._z = round(unwrap_if_numpy(z))

    @property
    def x(self) -> int:
        """
        Position of PointVolume on X-axis.

        :return: X of PointVolume
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            print(loc.x)
            # Output: 100
        """
        return self._x

    @property
    def y(self) -> int:
        """
        Position of PointVolume on Y-axis.

        :return: Y of PointVolume
        :rtype: :class:`int`

        :Usage example:

         .. code-block:: python

            print(loc.y)
            # Output: 200
        """
        return self._y

    @property
    def z(self) -> int:
        """
        Position of PointVolume on Z-axis.

        :return: Z of PointVolume
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            print(loc.z)
            # Output: 2
        """
        return self._z

    def to_json(self) -> Dict:
        """
        Convert the PointVolume to a json dict.

        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            loc_json = loc.to_json()
            print(loc_json)
            # Output: {
            #           "space_origin": [
            #                            200,
            #                            200,
            #                            100
            #                           ]
            #         }

        """

        packed_obj = {SPACE_ORIGIN: [self.x, self.y, self.z]}
        return packed_obj

    @classmethod
    def from_json(cls, packed_obj) -> PointVolume:
        """
        Convert a json dict to PointVolume.

        :param data: PointVolume in json format as a dict.
        :type data: dict
        :return: PointVolume object
        :rtype: :class:`PointVolume<PointVolume>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            loc_json = {
                "space_origin": [
                                200,
                                200,
                                100,
                                ]
                        }

            loc = sly.PointVolume.from_json(loc_json)
        """
        return cls(
            x=packed_obj["space_origin"][0],
            y=packed_obj["space_origin"][1],
            z=packed_obj["space_origin"][2],
        )


class Mask3D(Geometry):
    """
    Mask 3D geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`Mask3D<Mask3D>` object is immutable.

    :param data: Mask 3D mask data. Must be a numpy array with only 2 unique values: [0, 1] or [0, 255] or [False, True].
    :type data: np.ndarray
    :param sly_id: Mask 3D ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which Mask 3D belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created Mask 3D.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when Mask 3D was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when Mask 3D was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :raises: :class:`ValueError`, if data is not bool or no pixels set to True in data
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Create simple Mask 3D
        mask3d = np.zeros((3, 3, 3), dtype=np.bool_)
        mask3d[0:2, 0:2, 0:2] = True

        shape = sly.Mask3D(mask3d)

        print(shape.data)
        # Output:
        #    [[[ True  True False]
        #      [ True  True False]
        #      [False False False]]

        #     [[ True  True False]
        #      [ True  True False]
        #      [False False False]]

        #     [[False False False]
        #      [False False False]
        #      [False False False]]]
    """

    def __init__(
        self,
        data: np.ndarray,
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        super().__init__(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

        if not isinstance(data, np.ndarray):
            raise TypeError('Mask 3D "data" argument must be numpy array object!')

        data_dims = len(data.shape)
        if data_dims != 3:
            raise ValueError(
                f'Mask 3D "data" argument must be a 3-dimensional numpy array. Instead got {data_dims} dimensions'
            )

        if data.dtype != np.bool:
            if list(np.unique(data)) not in [[0, 1], [0, 255]]:
                raise ValueError(
                    f"Mask 3D mask data values must be one of: [0 1], [0 255], [False True]. Instead got {np.unique(data)}."
                )

            if list(np.unique(data)) == [0, 1]:
                data = np.array(data, dtype=bool)
            elif list(np.unique(data)) == [0, 255]:
                data = np.array(data / 255, dtype=bool)

        self.data = data
        self._space_origin = None
        self._space = None
        self._space_directions = None

    @staticmethod
    def geometry_name():
        """Return geometry name"""
        return "mask_3d"

    @staticmethod
    def from_file(figure, file_path: str):
        """
        Load figure geometry from file.

        :param figure: Spatial figure
        :type figure: VolumeFigure
        :param file_path: Path to nrrd file with data
        :type file_path: str
        """
        mask3d_data, mask3d_header = nrrd.read(file_path)
        figure.geometry.data = mask3d_data
        try:
            figure.geometry._space_origin = PointVolume(
                x=mask3d_header["space origin"][0],
                y=mask3d_header["space origin"][1],
                z=mask3d_header["space origin"][2],
            )
            figure.geometry._space = mask3d_header["space"]
            figure.geometry._space_directions = mask3d_header["space directions"]
        except KeyError as e:
            header_keys = ["'space'", "'space directions'", "'space origin'"]
            if str(e) in header_keys:
                logger.warning(
                    f"The Mask3D geometry for figure ID '{get_file_name(file_path)}' doesn't contain optional space attributes that have similar names to {', '.join(header_keys)}. To set the values for these attributes, you can use information from the Volume associated with this figure object."
                )
        path_without_filename = "/".join(file_path.split("/")[:-1])
        remove_dir(path_without_filename)

    @classmethod
    def create_from_file(cls, file_path: str) -> Mask3D:
        """
        Creates Mask3D geometry from file.

        :param file_path: Path to nrrd file with data
        :type file_path: str
        """
        mask3d_data, mask3d_header = nrrd.read(file_path)
        geometry = cls(data=mask3d_data)
        try:
            geometry._space_origin = PointVolume(
                x=mask3d_header["space origin"][0],
                y=mask3d_header["space origin"][1],
                z=mask3d_header["space origin"][2],
            )
            geometry._space = mask3d_header["space"]
            geometry._space_directions = mask3d_header["space directions"]
        except KeyError as e:
            header_keys = ["'space'", "'space directions'", "'space origin'"]
            if str(e) in header_keys:
                logger.warning(
                    f"The Mask3D geometry created from the file '{file_path}' doesn't contain optional space attributes that have similar names to {', '.join(header_keys)}. To set the values for these attributes, you can use information from the Volume associated with this figure object."
                )
        return geometry

    @classmethod
    def from_bytes(cls, geometry_bytes: bytes) -> Mask3D:
        """
        Create a Mask3D geometry object from bytes.

        :param geometry_bytes: NRRD file represented as bytes.
        :type geometry_bytes: bytes
        :return: A Mask3D geometry object.
        :rtype: Mask3D
        """
        with tempfile.NamedTemporaryFile(delete=True) as temp_file:
            temp_file.write(geometry_bytes)
            data_array, _ = nrrd.read(temp_file.name)

        return cls(data_array)

    def to_json(self) -> Dict:
        """
        Convert the Mask 3D to a json dict.

        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            mask = np.array([[[1 1 0]
                              [1 1 0]
                              [0 0 0]]
                             [[1 1 0]
                              [1 1 0]
                              [0 0 0]]
                             [[0 0 0]
                              [0 0 0]
                              [0 0 0]]], dtype=np.bool_)

            figure = sly.Mask3D(mask)
            figure_json = figure.to_json()
            print(json.dumps(figure_json, indent=4))
            # Output: {
            #    "mask_3d": {
            #        "data": "eJzrDPBz5+WS4mJgYOD19HAJAtLMIMwIInOeqf8BUmwBPiGuQPr///9Lb86/C2QxlgT5BTM4PLuRBuTwebo4hlTMSa44sKHhISMDuxpTYrr03F6gDIOnq5/LOqeEJgDM5ht6",
            #    },
            #    "shape": "mask_3d",
            #    "geometryType": "mask_3d"
            # }
        """
        res = {
            self._impl_json_class_name(): {
                DATA: self.data_2_base64(self.data),
            },
            GEOMETRY_SHAPE: self.name(),
            GEOMETRY_TYPE: self.name(),
        }

        if self._space_origin:
            res[f"{self._impl_json_class_name()}"][f"{SPACE_ORIGIN}"] = [
                self._space_origin.x,
                self._space_origin.y,
                self._space_origin.z,
            ]

        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, json_data: Dict) -> Mask3D:
        """
        Convert a json dict to Mask 3D.

        :param data: Mask in json format as a dict.
        :type data: dict
        :return: Mask3D object
        :rtype: :class:`Mask3D<Mask3D>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_json = {
                "mask_3d": {
                    "data": "eJzrDPBz5+WS4mJgYOD19HAJAtLMIMwIInOeqf8BUmwBPiGuQPr///9Lb86/C2QxlgT5BTM4PLuRBuTwebo4hlTMSa44sKHhISMDuxpTYrr03F6gDIOnq5/LOqeEJgDM5ht6",
                },
                "shape": "mask_3d",
                "geometryType": "mask_3d"
            }

            figure = sly.Mask3D.from_json(figure_json)
        """
        if json_data == {}:
            return cls(data=np.zeros((3, 3, 3), dtype=np.bool_))

        json_root_key = cls._impl_json_class_name()
        if json_root_key not in json_data:
            raise ValueError(
                "Data must contain {} field to create Mask3D object.".format(json_root_key)
            )

        if DATA not in json_data[json_root_key]:
            raise ValueError(
                "{} field must contain {} and {} fields to create Mask3D object.".format(
                    json_root_key, DATA
                )
            )

        data = cls.base64_2_data(json_data[json_root_key][DATA])

        labeler_login = json_data.get(LABELER_LOGIN, None)
        updated_at = json_data.get(UPDATED_AT, None)
        created_at = json_data.get(CREATED_AT, None)
        sly_id = json_data.get(ID, None)
        class_id = json_data.get(CLASS_ID, None)
        instance = cls(
            data=data.astype(np.bool_),
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        if SPACE_ORIGIN in json_data[json_root_key]:
            x, y, z = json_data[json_root_key][SPACE_ORIGIN]
            instance._space_origin = PointVolume(x=x, y=y, z=z)
        return instance

    @classmethod
    def _impl_json_class_name(cls):
        """_impl_json_class_name"""
        return MASK_3D

    @staticmethod
    def data_2_base64(data: np.ndarray) -> str:
        """
        Convert numpy array to base64 encoded string.

        :param mask: Bool numpy array.
        :type mask: np.ndarray
        :return: Base64 encoded string
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            import os
            import nrrd

            address = 'https://app.supervise.ly/'
            token = 'Your Supervisely API Token'
            api = sly.Api(address, token)

            meta_json = api.project.get_meta(PROJECT_ID)
            meta = sly.ProjectMeta.from_json(meta_json)

            ann_json = api.volume.annotation.download_bulk(DATASET_ID, [VOLUME_ID])

            figure_id = ann_json[0]["spatialFigures"][0]["id"]
            path_for_mesh = f"meshes/{figure_id}.nrrd"
            api.volume.figure.download_stl_meshes([figure_id], [path_for_mesh])

            mask3d_data, _ = nrrd.read(path_for_mesh)
            encoded_string = sly.Mask3D.data_2_base64(mask3d_data)

            print(encoded_string)
            # 'H4sIAGWoWmQC/zPWMdYxrmFkZAAiIIAz4AAAE56ciyEAAAA='
        """
        shape_str = ",".join(str(dim) for dim in data.shape)
        data_str = data.tostring().decode("utf-8")
        combined_str = f"{shape_str}|{data_str}"
        compressed_string = gzip.compress(combined_str.encode("utf-8"))
        encoded_string = base64.b64encode(compressed_string).decode("utf-8")
        return encoded_string

    @staticmethod
    def base64_2_data(encoded_string: str) -> np.ndarray:
        """
        Convert base64 encoded string to numpy array.

        :param s: Input base64 encoded string.
        :type s: str
        :return: Bool numpy array
        :rtype: :class:`np.ndarray`
        :Usage example:

         .. code-block:: python

              import supervisely as sly

              encoded_string = 'H4sIAGWoWmQC/zPWMdYxrmFkZAAiIIAz4AAAE56ciyEAAAA='
              figure_data = sly.Mask3D.base64_2_data(encoded_string)
              print(figure_data)
              # [[[1 1 0]
              #   [1 1 0]
              #   [0 0 0]]
              #  [[1 1 0]
              #   [1 1 0]
              #   [0 0 0]]
              #  [[0 0 0]
              #   [0 0 0]
              #   [0 0 0]]]
        """
        compressed_bytes = base64.b64decode(encoded_string)
        decompressed_string = gzip.decompress(compressed_bytes).decode("utf-8")
        shape_str, data_str = decompressed_string.split("|")
        shape = tuple(int(dim) for dim in shape_str.split(","))
        data_bytes = data_str.encode("utf-8")
        try:
            data = np.frombuffer(data_bytes, dtype=np.uint8).reshape(shape)
        except ValueError:
            logger.warn(
                "Can't reshape array with 'dtype=np.uint8'. Will try to automatically convert 'dtype=np.int16' to 'np.uint8' and reshape"
            )
            data = np.frombuffer(data_bytes, dtype=np.int16)
            data = np.clip(data, 0, 1).astype(np.uint8)
            data = data.reshape(shape)
            logger.debug("Converted successfully!")
        return data

    def add_mask_2d(
        self,
        mask_2d: np.ndarray,
        plane_name: Literal["axial", "sagittal", "coronal"],
        slice_index: int,
        origin: Optional[List[int]] = None,
    ):
        """
        Draw a 2D mask on a 3D Mask.

        :param mask_2d: 2D array with a flat mask.
        :type mask_2d: np.ndarray
        :param plane_name: Name of the plane: "axial", "sagittal", "coronal".
        :type plane_name: str
        :param slice_index: Slice index of the volume figure.
        :type slice_index: int
        :param origin: (row, col) position. The top-left corner of the mask is located on the specified slice (optional).
        :type origin: Optional[List[int]], NoneType
        """

        from supervisely.volume_annotation.plane import Plane

        Plane.validate_name(plane_name)

        mask_2d = np.fliplr(mask_2d)
        mask_2d = np.rot90(mask_2d, 1, (1, 0))

        if plane_name == Plane.AXIAL:
            new_shape = self.data.shape[:2]
        elif plane_name == Plane.SAGITTAL:
            new_shape = self.data.shape[1:]
        elif plane_name == Plane.CORONAL:
            new_shape = self.data.shape[::2]

        if origin:
            x, y = origin
            new_mask = np.zeros(new_shape, dtype=mask_2d.dtype)
            new_mask[x : x + mask_2d.shape[0], y : y + mask_2d.shape[1]] = mask_2d

        if plane_name == Plane.AXIAL:
            self.data[:, :, slice_index] = new_mask
        elif plane_name == Plane.SAGITTAL:
            self.data[slice_index, :, :] = new_mask
        elif plane_name == Plane.CORONAL:
            self.data[:, slice_index, :] = new_mask



from typing import List, Optional
from supervisely.app.widgets import Widget
from supervisely.app.content import StateJson, DataJson


class Timeline(Widget):
    class Routes:
        SEGMENT_SELECTED = "segment_selected_cb"
        CLICK = "click"

    def __init__(
        self,
        frames_count: int,
        intervals: List,
        colors: List,
        height: Optional[str] = "30px",
        pointer_color: Optional[str] = None,
        tooltip_content: Optional[Widget] = None,
        widget_id: str = None,
    ):
        self._frames_count = frames_count
        self._intervals = intervals
        self._colors = colors
        if len(self._intervals) != len(self._colors):
            raise ValueError("Intervals and colors lists must be of the same length")
        for interval in self._intervals:
            if interval[0] < 0 or interval[1] > self._frames_count:
                raise ValueError("Interval is out of bounds")
        self._height = height
        if pointer_color is None:
            pointer_color = "rgba(151, 151, 151, 1)"
        self._pointer_color = pointer_color
        self._tooltip_content = tooltip_content
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "framesCount": self._frames_count,
            "intervals": self._intervals,
            "colors": self._colors,
            "options": {
                "height": self._height,
                "pointerColor": self._pointer_color,
            },
        }

    def get_json_state(self):
        return {
            "pointer": 1,
            "selectedSegment": None,
        }

    def set_pointer(self, pointer):
        StateJson()[self.widget_id]["pointer"] = pointer
        StateJson().send_changes()

    def get_pointer(self):
        return StateJson()[self.widget_id]["pointer"]

    def get_selected_segment(self):
        return StateJson()[self.widget_id]["selectedSegment"]

    def set(self, frames_count, intervals, colors):
        if len(intervals) != len(colors):
            raise ValueError("Intervals and colors lists must be of the same length")
        for interval in intervals:
            if interval[0] < 0 or interval[1] > frames_count:
                raise ValueError("Interval is out of bounds")
        self._frames_count = frames_count
        self._intervals = intervals
        self._colors = colors
        DataJson()[self.widget_id] = self.get_json_data()
        DataJson().send_changes()

    def segment_selected(self, func):
        route_path = self.get_route_path(Timeline.Routes.SEGMENT_SELECTED)
        server = self._sly_app.get_server()
        self._segment_selected_handled = True

        @server.post(route_path)
        def _inner():
            selected_segment = StateJson()[self.widget_id]["selectedSegment"]
            func(selected_segment)

        return _inner

    def click(self, func):
        route_path = self.get_route_path(Timeline.Routes.CLICK)
        server = self._sly_app.get_server()
        self._click_handled = True

        @server.post(route_path)
        def _inner():
            pointer = StateJson()[self.widget_id]["pointer"]
            func(pointer)

        return _inner
